use crate::{checkout, init};
use std::{
    path::PathBuf,
    process::{exit, Command, Stdio},
};

/// Check if .borg repository and results folder exists
/// # Return
/// - ` borg_folder`: Path to .borg folder
/// - `results_folder`: Path to the result folder
pub fn check_path() -> (PathBuf, PathBuf) {
    let (borg_folder, results_folder) = init::get_borg_folder();
    if !borg_folder.is_dir() {
        eprintln!("{} folder not found !", borg_folder.to_str().unwrap());
        exit(2)
    }
    (borg_folder, results_folder)
}

/// Get the current commit name of a git repository
///
/// # Return:
/// * A string corresponding to the current commit id
pub fn get_current_commit() -> String {
    let output = Command::new("git")
        .arg("rev-parse")
        .arg("--verify")
        .arg("HEAD")
        .output()
        .unwrap();
    match output.status.code().unwrap() {
        0 => (),
        num => {
            eprintln!("{}", String::from_utf8(output.stderr).unwrap());
            exit(num);
        }
    }
    let mut commit = String::from_utf8(output.stdout).unwrap();
    commit.pop();
    commit
}

/// function that deletes commit if needed
pub fn delete_commit(commit: &str, borg_path: &PathBuf) {
    let output = Command::new("borg")
        .arg("delete")
        .arg(format!("{}::{}", borg_path.to_str().unwrap(), commit))
        .output()
        .unwrap();
    match output.status.code().unwrap() {
        0 => (),
        num => {
            eprintln!("{}", String::from_utf8(output.stderr).unwrap());
            exit(num);
        }
    }
}

/// Checks if .borgignore file exists in the project folder
///
/// # Arguments:
/// * borg_folder: Path to .borg folder
///
/// # Return:
/// A tuple containing:
/// * A boolean indicating if .borgignore file exits
/// * The path to this borgignore file
pub fn check_if_borgignore_exists(borg_folder: &PathBuf) -> (bool, PathBuf) {
    let mut project_folder = borg_folder.parent().unwrap().to_path_buf();
    project_folder.push(".borgignore");
    (project_folder.is_file(), project_folder)
}

/// Function that checks if the archive with the name of the current commit id
/// exits in borg repository
pub fn is_a_commit(borg_path: &PathBuf, commit_id: &str) -> bool {
    // Trying to create an archive with the current git commit id
    let output = Command::new("borg")
        .arg("list")
        .arg(format!("{}::{}", borg_path.to_str().unwrap(), commit_id))
        .output()
        .unwrap();
    match output.status.code().unwrap() {
        0 => true,
        _ => false,
    }
}

/// Revert the result folder back to the current commit archive
pub fn revert_commit() {
    let commit_id = get_current_commit();
    let (borg_path, _) = check_path();
    if is_a_commit(&borg_path, &commit_id) {
        checkout::checkout("hard");
    } else {
        eprintln!("Cannot revert the content of the results folder previously saved for that commit: No archive found for this commit !");
        exit(17);
    }
}

/// Create a commit of the results folder named as the current git commit id
pub fn commit(compression: String, mut commit_id: String, update: bool) {
    let (borg_folder, results_folder) = check_path();
    let (has_ignore, borgignore) = check_if_borgignore_exists(&borg_folder);
    if commit_id == String::from("") {
        commit_id = get_current_commit();
    }
    if update {
        delete_commit(&commit_id, &borg_folder);
    }

    let mut margs: Vec<&str> = vec![
        "create",
        "--stats",
        "--progress",
        "--compression",
        &compression,
    ];
    if has_ignore {
        margs.push("--patterns-from");
        margs.push(borgignore.to_str().unwrap());
    }
    let cmd_part = format!("{}::{}", borg_folder.to_str().unwrap(), commit_id);
    margs.push(&cmd_part);
    margs.push(results_folder.to_str().unwrap());
    let mut output = Command::new("borg")
        .args(margs)
        .stdout(Stdio::piped())
        .spawn()
        .unwrap_or_else(|e| {
            eprintln!("An error occured during borg create ! {}", e);
            exit(8);
        });
    let ecode = output.wait().expect("error");
    match ecode.code().unwrap() {
        0 => (),
        num => {
            eprintln!("{}", ecode.to_string());
            exit(num);
        }
    }
}
