use crate::commit;
use crate::config_structure::{Config, GblkConfig, RemoteConfig};
use crate::prune::{new_prune, prune_launcher};
use clap::Args;
use colored::Colorize;
use home;
use std::{path::PathBuf, process::exit};
use toml;

#[derive(Debug, Args)]
pub(crate) struct PartialPrune {
    /// Do not change the repository
    #[clap(
        short = 'n',
        long = "dry-run",
        takes_value = false,
        help_heading = "Filtering options",
        display_order = 1
    )]
    pub(crate) dry_run: bool,
    /// Output verbose list of archive
    #[clap(
        long = "--list",
        takes_value = false,
        help_heading = "Filtering options",
        display_order = 2
    )]
    pub(crate) list: bool,
    /// Print statistics for the deleted archive
    #[clap(
        short = 's',
        long = "stats",
        takes_value = false,
        help_heading = "Filtering options",
        display_order = 3
    )]
    pub(crate) stats: bool,
    /// Force deletion of corrupted archives, use `--force --force` in case `--force` does not work.
    #[clap(
        long,
        parse(from_occurrences),
        takes_value = false,
        help_heading = "Filtering options",
        display_order = 4
    )]
    pub(crate) force: usize,
}

/// Get the config file inside borg folder
/// # Return
/// A pathbuf variable containing the location of borg config file for the
/// current project
pub fn get_borgconfig() -> Option<PathBuf> {
    let (mut config_path, _) = commit::check_path();
    config_path.push(".gblkconfig");
    Some(config_path)
}

/// Get the global configuration for gblk
/// # Return
/// - The path where the global location of the gblk configuration should be set
pub fn get_global_config() -> Option<PathBuf> {
    let global_config = match home::home_dir() {
        Some(path) => {
            let mut tmp = path;
            tmp.push(".gblkconfig");
            Some(tmp)
        }
        None => None,
    };
    global_config
}

/// Get the configuration file of interest
/// # Argument
/// - global: A bolean indicating whether to get the global configuration or
///   only the local one
/// # Return
/// The path containing the local of global configuration
pub(crate) fn get_config(global: bool) -> PathBuf {
    if global {
        match get_global_config() {
            Some(path) => path,
            None => {
                eprintln!("{} config not found !", "global".red());
                exit(0);
            }
        }
    } else {
        match get_borgconfig() {
            Some(path) => path,
            None => {
                eprintln!("{} config not found !", "local".red());
                exit(0);
            }
        }
    }
}

/// Return the parsed config of interest
/// # Arguments
/// - global: A bolean indicating whether to get the global configuration or
///   only the local one
/// # Return
/// The parsed toml file
pub fn parse_toml(global: bool) -> Config {
    let config_file = get_config(global);
    let rcontent = std::fs::read_to_string(&config_file);
    let content = match rcontent {
        Err(e) => match e.raw_os_error() {
            Some(2) => String::from(""),
            _ => {
                eprintln!("Unable to read {} - {}", &config_file.to_str().unwrap(), e);
                exit(101);
            }
        },
        Ok(s) => s,
    };
    if content.is_empty() {
        if global {
            return Config::new(global);
        } else {
            eprintln!(
                "{}: Loccal config file should not be empty !",
                "error".red()
            );
            exit(1);
        }
    }
    let config_str: Config = toml::from_str(&content).unwrap();
    config_str.check_project_type(global, &config_file);
    config_str
}

/// Function that show the current gblk config of the borg folder
/// # Arguments
/// - global: A bolean indicating if we whant to show the config of the current
///   profile or the global configuration
pub fn show(global: bool) -> () {
    let config = parse_toml(global);
    let kind = if global { "global" } else { "local" };
    let gblk_config: GblkConfig = match config.gblk_prune {
        Some(data) => data,
        None => {
            println!(
                "{} {} {}",
                "No".yellow(),
                kind.green(),
                "configuration defined".yellow()
            );
            exit(0);
        }
    };
    let gblk_str = toml::to_string_pretty(&gblk_config).unwrap();
    println!("{}", gblk_str);
}

/// Function that update the current gblk config file for the project folder
/// # Arguments
/// - `key` : The key to update
/// - `value`: The value of the key to update
/// # Return
/// The updated prune keys
fn update_val(input: &mut GblkConfig, key: &str, value: &str) -> () {
    match key {
        "save_space" | "save-space" => {
            input.save_space = Some(str::parse::<bool>(value).unwrap());
        }
        "keep_within" | "keep-within" => {
            input.keep_within = Some(value.to_owned());
        }
        "keep_last" | "keep-last" => {
            input.keep_last = Some(str::parse::<usize>(value).unwrap());
        }
        "keep_minutely" | "keep-minutely" => {
            input.keep_minutely = Some(str::parse::<usize>(value).unwrap());
        }
        "keep_hourly" | "keep-hourly" => {
            input.keep_hourly = Some(str::parse::<usize>(value).unwrap());
        }
        "keep_daily" | "keep-daily" => {
            input.keep_daily = Some(str::parse::<usize>(value).unwrap());
        }
        "keep_weekly" | "keep-weekly" => {
            input.keep_weekly = Some(str::parse::<usize>(value).unwrap());
        }
        "keep_monthly" | "keep-monthly" => {
            input.keep_monthly = Some(str::parse::<usize>(value).unwrap());
        }
        "keep_yearly" | "keep-yearly" => {
            input.keep_yearly = Some(str::parse::<usize>(value).unwrap());
        }
        "prefix" => {
            input.prefix = Some(value.to_owned());
        }
        "glob_archives" | "glob-archives" => {
            input.glob_archives = Some(value.to_owned());
        }
        &_ => {
            eprintln!(
                "{} {} '{}' {}\n{}\n- {}\n- {}",
                "error:".red(),
                "The key",
                key,
                "doesn't exist !",
                "choose from:",
                "keep_within, keep_last, keep_minutely, keep_hourly, keep_daily".blue(),
                "keep_weekly, keep_monthly, keep_yearly, prefix, glob_archives, save_space".blue()
            );
            exit(1);
        }
    }
}

/// Function that checks if the parameter input param is already defined
/// # Description:
/// If input_param is already defined, returns None but if not, return an error
/// # Arguments:
/// - input_param: A parameter of the gblk config structure
/// - key: The key to remove
/// # Return
/// None if input_param is defined, error else
fn check_and_warn<T>(input_param: &Option<T>, key: &str) -> Option<T> {
    match input_param {
        None => {
            eprintln!(
                "{} {} {} {}",
                "error:".red(),
                "The key",
                key,
                "is not defined !"
            );
            exit(103);
        }
        Some(_) => None,
    }
}

/// Function that remove a key from the current gblk config file for the project folder
/// # Arguments
/// - `key` : The key to remove
/// # Return
/// The updated prune keys
fn remove_val(input: &mut GblkConfig, key: &str) -> () {
    match key {
        "save_space" | "save-space" => {
            input.save_space = check_and_warn(&input.save_space, &key);
        }
        "keep_within" | "keep-within" => {
            input.keep_within = check_and_warn(&input.keep_within, &key);
        }
        "keep_last" | "keep-last" => {
            input.keep_last = check_and_warn(&input.keep_last, &key);
        }
        "keep_minutely" | "keep-minutely" => {
            input.keep_minutely = check_and_warn(&input.keep_minutely, &key);
        }
        "keep_hourly" | "keep-hourly" => {
            input.keep_hourly = check_and_warn(&input.keep_hourly, &key);
        }
        "keep_daily" | "keep-daily" => {
            input.keep_daily = check_and_warn(&input.keep_daily, &key);
        }
        "keep_weekly" | "keep-weekly" => {
            input.keep_weekly = check_and_warn(&input.keep_weekly, &key);
        }
        "keep_monthly" | "keep-monthly" => {
            input.keep_monthly = check_and_warn(&input.keep_monthly, &key);
        }
        "keep_yearly" | "keep-yearly" => {
            input.keep_yearly = check_and_warn(&input.keep_yearly, &key);
        }
        "prefix" => {
            input.prefix = check_and_warn(&input.prefix, &key);
        }
        "glob_archives" | "glob-archives" => {
            input.glob_archives = check_and_warn(&input.glob_archives, &key);
        }
        &_ => {
            eprintln!(
                "{} {} '{}' {}\n{}\n- {}\n- {}",
                "error:".red(),
                "The key",
                key,
                "doesn't exist !",
                "choose from:",
                "keep_within, keep_last, keep_minutely, keep_hourly, keep_daily".blue(),
                "keep_weekly, keep_monthly, keep_yearly, prefix, glob_archives, save_space".blue()
            );
            exit(1);
        }
    }
}

/// Get the object conrresponding to gblk prune config
/// # Arguments
/// - full_config: The entire configuration file
/// # Return
/// The configuration for gblk prune object
fn get_gblkconfig(full_config: &Config) -> GblkConfig {
    let gblk_config: GblkConfig = match &full_config.gblk_prune {
        Some(data) => data.clone(),
        None => GblkConfig {
            save_space: None,
            keep_within: None,
            keep_last: None,
            keep_minutely: None,
            keep_hourly: None,
            keep_daily: None,
            keep_weekly: None,
            keep_monthly: None,
            keep_yearly: None,
            prefix: None,
            glob_archives: None,
        },
    };
    gblk_config
}

/// Function that turn toml into string
/// # Argument
/// - config: The full toml file parsed
/// # Return
/// The string corresponding to toml file
fn toml_to_string(config: &Config) -> String {
    let gblk_str = toml::to_string_pretty(&config).unwrap();
    gblk_str
}

/// Function used to write a configuration file
/// # Arguments
/// - `config` The structure representing the configuration file to write
/// - `global`: A boolean indicating whether or not to write the configuration file
pub fn write_config(config: &Config, global: bool) {
    let gblk_str = toml_to_string(&config);
    let config_file = get_config(global);
    let kind = if global { "global" } else { "local" };
    match std::fs::write(&config_file, gblk_str) {
        Err(_) => {
            eprintln!(
                "{} {} {} {} {} {}",
                "error:".red(),
                "Unable to write the",
                kind.green(),
                "config file",
                config_file.to_str().unwrap(),
                "!"
            );
            exit(102);
        }
        Ok(_) => (),
    };
}

/// Create a local config file
pub fn create_local_config() {
    let conf = Config::new(false);
    write_config(&conf, false);
}

/// Create a named local config file
/// # Arguments
/// - `name`: The name of the config file
pub fn create_named_local_config(name: &str) {
    let conf = Config::new_named(name);
    write_config(&conf, false);
}

/// Function that update the current gblk config file for the project folder
/// # Arguments
/// - `key` : The key to update
/// - `value`: The value of the key to update
/// - `global: A bolean indicating whether to update the  global or the local configuration
pub fn update_config(key: &str, value: &str, global: bool) -> () {
    let mut config = parse_toml(global);
    let mut gblk_config: GblkConfig = get_gblkconfig(&config);
    update_val(&mut gblk_config, key, value);
    config.gblk_prune = Some(gblk_config);
    config.gblk_remote = Some(vec![
        RemoteConfig {
            name: String::from("lol"),
            url: String::from("test"),
        },
        RemoteConfig {
            name: String::from("lol2"),
            url: String::from("test2"),
        },
    ]);
    write_config(&config, global);
}

/// Function that remove the current gblk config file for the project folder
/// for a given arguments
/// # Arguments
/// - `key` : The key to update
/// - `global`: A bolean indicating whether to remove a global or a local
///   configuration file
pub fn remove_config(key: &str, global: bool) -> () {
    let mut config = parse_toml(global);
    let mut gblk_config: GblkConfig = get_gblkconfig(&config);
    remove_val(&mut gblk_config, key);
    config.gblk_prune = if gblk_config.empty() {
        None
    } else {
        Some(gblk_config)
    };
    let gblk_str = toml_to_string(&config);
    let config_file = get_config(global);
    let kind = if global { "global" } else { "local" };
    match std::fs::write(&config_file, gblk_str) {
        Err(_) => {
            eprintln!(
                "{} {} {} {} {} {}",
                "error:".red(),
                "Unable to write the",
                kind.green(),
                "config file",
                config_file.to_str().unwrap(),
                "!"
            );
            exit(102);
        }
        Ok(_) => (),
    };
}

/// Get the name of the current project
pub(crate) fn get_project_name() -> String {
    let config = parse_toml(false);
    match config.get_name() {
        None => {
            eprintln!("{}: Unable to get the project name !", "error".red());
            exit(1);
        }
        Some(name) => name,
    }
}

/// Function that recover the local or global gblk config
/// # Description
/// This function return the parsed local gblk config file if it exists. If
/// this config is not defined, then it will return tje global gblk config file
fn parse_local_or_global_config() -> GblkConfig {
    let config = parse_toml(false);
    let gblk_config: GblkConfig = get_gblkconfig(&config);
    if gblk_config.empty() {
        let config = parse_toml(true);
        let gblk_config: GblkConfig = get_gblkconfig(&config);
        return gblk_config;
    }
    gblk_config
}

/// Execute a prune based on gblk configuration
/// # Arguments
/// - `config_prune`: contains partial configuration for prune that cannot be
///   used in the configuration file
/// - `global`: a boolean indicating whether to prune using
pub(crate) fn launch_config_prune(config_prune: PartialPrune) {
    let gblk_config = parse_local_or_global_config();
    if gblk_config.empty() {
        eprintln!(
            "{} No global and local configuration defined for pruning!",
            "error:".red()
        );
        exit(104);
    }
    let prune_obj = new_prune(gblk_config, config_prune);
    prune_launcher(prune_obj);
}
