use std::fs;
use std::fs::File;
use std::fs::OpenOptions;
use std::io::Write;
use std::os::unix::prelude::OpenOptionsExt;
use std::{
    path::PathBuf,
    process::{exit, Command},
};

/// Function used to get the git hooks repository
fn get_hooks_folder() -> PathBuf {
    let output = Command::new("git")
        .arg("rev-parse")
        .arg("--show-cdup")
        .output()
        .unwrap();
    match output.status.code().unwrap() {
        128 => {
            eprintln!("Not a git repository !");
            exit(128);
        }
        127 => {
            eprintln!("Git not found !");
            exit(127);
        }
        0 => (),
        num => {
            eprintln!("{}", String::from_utf8(output.stderr).unwrap());
            exit(num)
        }
    }
    let mut string_path = String::from_utf8(output.stdout).unwrap();
    if string_path.ends_with('\n') {
        string_path.pop();
    }
    let mut p = PathBuf::from(&string_path);
    p.push(".git");
    p.push("hooks");
    if !p.is_dir() {
        eprintln!("Folder {} not found !", p.to_str().unwrap());
        exit(5);
    }
    p
}

/// Function used to create a file
///
/// # Arguments:
/// * `folder`: The folder where the file will be created
/// * `file_name`: The file name of the file to create
/// * `content`: The content of the file to create
fn create_file(folder: &PathBuf, file_name: &str, content: &str) {
    let mut hfile = folder.to_owned();
    hfile.push(file_name);
    if hfile.is_file() {
        eprintln!(
            "Warning: The file {} already exists. Skipping...",
            hfile.to_str().unwrap()
        );
        return ();
    }
    let mut file = OpenOptions::new()
        .create_new(true)
        .write(true)
        .mode(0o755)
        .open(&hfile)
        .unwrap();
    File::create(&hfile).unwrap_or_else(|e| {
        eprintln!(
            "Unable to create the file {}.\n{}",
            hfile.to_str().unwrap(),
            e
        );
        exit(6);
    });
    write!(file, "{}", content).unwrap_or_else(|e| {
        eprintln!(
            "Unable to write the file {}.\n{}",
            hfile.to_str().unwrap(),
            e
        );
        exit(7);
    });
}

/// Add a gblkco alias in `.git/config` file at the end to ensure a quiet
/// Checkout and a quiet cancel of this checkout
fn update_config_file() {
    let git_hooks = get_hooks_folder();
    let mut git_config = git_hooks.parent().unwrap().to_path_buf();
    git_config.push("config");
    if !git_config.is_file() {
        eprintln!("Error: The file .git/config wasn't found");
        exit(19);
    }
    let content = fs::read_to_string(&git_config).unwrap();
    let alias = format!(
        "{}\n\t{}\n\t{}",
        "co = checkout -q",
        "conh = -c core.hooksPath=/dev/null checkout",
        "cnh = -c core.hooksPath=/dev/null commit"
    );
    if !content.contains(&alias) {
        let mut file = OpenOptions::new()
            .write(true)
            .append(true)
            .open(&git_config)
            .unwrap();
        writeln!(file, "[alias]\n\t{}", alias).unwrap();
    }
}

/// Create 2 files in `.git/hooks` folder
///
/// 1. Create post-commit hook
/// 2. Create post-checkout hook
///
/// # Arguments
/// * `compression`: The compression that will automatically be used to save the results folder (no, lz4, zstd, zlib or lzma) after a git commit
/// * `mode`: `mode`: The checkout mode used by gblk automatically after a git checkout: soft or hard.
pub fn create_hooks(compression: &str, mode: &str) {
    update_config_file();
    let git_folder = get_hooks_folder();
    let post_commit_cmd = format!("gblk commit --compression {}", &compression);
    let post_commit_cmd = post_commit_cmd.as_str();
    let post_co_cmd = format!("res=$(git reflog show | grep 'HEAD@{{0}}' | tr ' ' '\\n')\n\
    arr=(${{res// / }})\n\
    original_commit=${{arr[5]}}\n\
    target_commit=${{arr[7]}}\n\
    \n\
    if [ \"${{original_commit}}\" != \"${{target_commit}}\" ]; then\n\
    \tgit -c core.hooksPath=/dev/null co ${{original_commit}} && gblk pre-co && git -c core.hooksPath=/dev/null checkout ${{target_commit}} && gblk checkout --mode {}\n\
    fi
    ", mode);
    let post_co_cmd = post_co_cmd.as_str();
    let v = vec![post_commit_cmd, post_co_cmd];
    let file_name = vec!["post-commit", "post-checkout"];
    for (c_cmd, fname) in v.iter().zip(file_name.iter()) {
        let content = format!(
            "#!/bin/bash \n\
        \n\
        {}
        \n\
        ",
            c_cmd
        );
        create_file(&git_folder, fname, &content);
    }
}

/// Delete 2 hooks files in `.git/hooks` folder
///
/// 1. Create post-commit hook
/// 2. Create post-checkout hook
pub fn delete_hooks() {
    let git_folder = get_hooks_folder();
    let file_name = vec!["post-commit", "post-checkout"];
    for cfile in file_name {
        let mut hfile = git_folder.to_owned();
        hfile.push(cfile);
        let cmd = format!("rm {}", hfile.to_str().unwrap());
        let output = Command::new("sh").arg("-c").arg(cmd).output().unwrap();
        match output.status.code().unwrap() {
            0 => (),
            num => {
                eprintln!("{}", String::from_utf8(output.stderr).unwrap());
                exit(num)
            }
        }
    }
}
