use colored::Colorize;

use crate::{commit, configt, create_hooks};
use std::io::Write;
use std::path::PathBuf;
use std::process::{exit, Command};

/// Function used to get borg repository folder and results repository only if
/// the project is inside a git repository
pub fn get_borg_folder() -> (PathBuf, PathBuf) {
    let output = Command::new("git")
        .arg("rev-parse")
        .arg("--show-cdup")
        .output()
        .unwrap();
    match output.status.code().unwrap() {
        128 => {
            eprintln!("Not a git repository !");
            exit(128);
        }
        127 => {
            eprintln!("Git not found !");
            exit(127);
        }
        0 => (),
        num => {
            eprintln!("{}", String::from_utf8(output.stderr).unwrap());
            exit(num)
        }
    }
    let mut string_path = String::from_utf8(output.stdout).unwrap();
    if string_path.ends_with('\n') {
        string_path.pop();
    }
    let mut p = PathBuf::from(&string_path);
    let mut results = p.clone();
    p.push(".borg");
    results.push("results");
    if !results.is_dir() {
        eprintln!("{} not found !", results.to_str().unwrap());
        exit(1);
    }
    (p, results)
}

/// Creation of a borg repository in the same directory as the .git repository
fn init_repository() {
    let (borg_path, _) = get_borg_folder();
    let output = Command::new("borg")
        .arg("init")
        .arg("--encryption=none")
        .arg(&borg_path)
        .output()
        .unwrap();
    match output.status.code().unwrap() {
        128 => {
            eprintln!("borg not found !");
            exit(128)
        }
        0 => {
            println!(
                "borg repository initialised at {}",
                borg_path.to_str().unwrap()
            );
        }
        num => {
            eprintln!("{}", String::from_utf8(output.stderr).unwrap());
            exit(num);
        }
    }
}

/// Function that returns the path to the .gitignore and restuls/.gitignore file
/// # Return
/// The path to the .gitignore file and results/.gitignore of the project
fn get_gitignore_file() -> (PathBuf, PathBuf) {
    let (borg_folder, results) = commit::check_path();
    let borg_folder = borg_folder.canonicalize().unwrap();
    let mut results = results.canonicalize().unwrap();
    let mut gitignore = borg_folder.parent().unwrap().to_path_buf();
    gitignore.push(".gitignore");
    results.push(".gitignore");
    (gitignore, results)
}

/// This function update the .gitignore file with folder created by gblk
/// # Arguments
/// - ` gitignore_file` : Path to the gitignore file
/// - `ignore`: The string to add in gitignore file
fn update_gitignore(gitignore_file: &PathBuf, ignore: &str) -> () {
    if !gitignore_file.is_file() {
        std::fs::File::create(gitignore_file).expect(&format!(
            "{}: Unable to create {} file",
            "error".red(),
            &gitignore_file.display()
        ));
    }
    let mut giti = std::fs::OpenOptions::new()
        .append(true)
        .open(&gitignore_file)
        .expect(&format!(
            "{}: Unable to open {} file",
            "error".red(),
            &gitignore_file.display()
        ));
    giti.write_all(ignore.as_bytes()).expect(&format!(
        "{}: Unable to write {} file",
        "error".red(),
        &gitignore_file.display()
    ));
}

/// Update .gitignore and results/.gitignore
pub(crate) fn update_gitignores() {
    let (gitignore, giti_res) = get_gitignore_file();
    update_gitignore(&gitignore, ".borg\n.mount\n.tmp\n");
    update_gitignore(&giti_res, "*\n!.gitignore");
}

/// Creation of a borg repository and creation of hooks if needed inside `.git/hooks`
/// folder
/// # Arguments:
/// * `hooks`: a boolean indicating wether or not to create hooks
/// * `compression`: The compression used automatically after each commit. This parameter is only used if hooks is true.
/// * `mode`: The checkout mode used by gblk automatically after a git checkout: soft or hard.
pub fn init_and_hook(hooks: bool, compression: &str, mode: &str) {
    init_repository();
    configt::create_local_config();
    update_gitignores();
    if hooks {
        create_hooks::create_hooks(compression, mode);
    }
}
