use crate::commit;
use crate::mount::file_diff;
use crate::push;
use colored::Colorize;
use std::path::PathBuf;
use std::process::{exit, Command, Stdio};

/// Function that gets the backup borg folder path used to restore or .borg
/// folder in case that the pull command fails
///
/// # Arguments
/// - ` borg_folder` : Path to .brg folder of the current project
/// # Return
/// backup borg folder path used to restore or .borg
/// folder in case that the pull command fails. It will be located in
/// .tmp/<PROJECT_DIR>_bkp folder. Where <PROJECT_DIR> is the folder name
/// containging our project
pub(crate) fn get_savefolder(borg_folder: &PathBuf) -> PathBuf {
    let mut save_dir = file_diff::get_tmp_folder(borg_folder);
    let folder_name = borg_folder
        .parent()
        .unwrap()
        .file_name()
        .unwrap()
        .to_str()
        .unwrap();
    save_dir.push(format!("{}_bkp", folder_name));
    save_dir
}

/// Function that removes the backup folder (`save_dir`) in the tmp directory if it exists
/// # Argument
/// - `save_dir`: The backup folder inside the .tmp directory
fn check_and_remove_dir(save_dir: &PathBuf) {
    if !save_dir.is_dir() {
        return ();
    }
    let anwser = push::user_question(&format!(
        "The backup folder {} already exists, Erase it with you're current .borg folder for backup",
        save_dir.display()
    ));
    if !anwser {
        println!("{}: Exiting !", "info".blue());
        exit(0);
    }
    match std::fs::remove_dir_all(save_dir) {
        Err(e) => {
            eprintln!(
                "{}: Unable to remove {}. {}",
                "error".red(),
                save_dir.display(),
                e
            );
            exit(1);
        }
        Ok(_) => (),
    };
}

/// Function that will remove almost all the content of the .borg folder
/// it only keeps .gblkconfig if it exists
/// # Arguments
/// - `borg_folder` : path to the .borg folder
fn save_and_clean_borg_folder(borg_folder: &PathBuf) {
    let save_dir = get_savefolder(borg_folder);
    println!(
        "{}: Saving current .borg repositoy into {} for backup.\n    - If everything works, you can remove it with the command {}.\n    - If nothing works anymore, you can restore it with {}",
        "info".blue(),
        &save_dir.display(),
        "gblk clean".green(),
        "gblk restore".green()
    );
    println!("{}", &save_dir.display());
    // check if the folder exits, removes it with the user permission
    check_and_remove_dir(&save_dir);
    match std::fs::create_dir(&save_dir) {
        Err(e) => {
            eprintln!(
                "{}: Unable to create {}. {}",
                "error".red(),
                &save_dir.display(),
                e
            );
            exit(1);
        }
        Ok(_) => (),
    }
    let paths = std::fs::read_dir(borg_folder).unwrap();
    for p in paths {
        let pa = p.unwrap().path();
        let name = pa.file_name().unwrap().to_str().unwrap();
        let mut destination = save_dir.to_owned();
        destination.push(name);
        match name {
            ".gblkconfig" | "config" => {
                std::fs::copy(&pa, &destination).expect(&format!(
                    "{}: Unable to copy {} to {}. Your .borg folder may be corrupted, you can restor it wiht the command {}",
                    "error".red(),
                    &pa.display(),
                    &destination.display(),
                    "gblk restore".green()
                ));
            }
            _ => {
                std::fs::rename(&pa, &destination).expect(&format!(
                    "{}: Unable to move {} to {} Your .borg folder may be corrupted, you can restor it wiht the command {}",
                    "error".red(),
                    &pa.display(),
                    &destination.display(),
                    "gblk restore".green()
                ));
            }
        }
    }
}

/// Function that check if `path_str` is an existing directory. If it exists,
/// return it else raise an error
/// # Arguments
/// - `path_str`: A path in string form
/// # Return
/// The pathbuf object of `path_str`
fn check_ok(path_str: &str) -> PathBuf {
    let path = PathBuf::from(path_str);
    if !path.is_dir() {
        eprintln!(
            "{}: {} is not a directory !",
            "error:".red(),
            path.display()
        );
        exit(1)
    } else {
        return path;
    }
}

/// Get the borg base directory
///
/// # Description
/// Get the directory defined with BORG_BASE_DIR env variable or the directory
/// defined in HOME variable
///
/// # Return
/// The borg base directory
fn get_base_dir() -> PathBuf {
    let base_dir = std::env::var("BORG_BASE_DIR");
    match base_dir {
        Ok(dir) => check_ok(&dir),
        Err(_) => {
            let base_dir = std::env::var("HOME");
            match base_dir {
                Ok(dir) => check_ok(&dir),
                Err(_) => {
                    eprintln!("{}: HOME environment variable not set !", "error".red());
                    exit(1);
                }
            }
        }
    }
}

/// Get the borg config directory
///
/// # Description
/// Get the directory defined with BORG_BASE_DIR env variable or the directory
/// defined in HOME variable
///
/// # Return
/// The borg base directory
fn get_config_dir() -> PathBuf {
    let mut config = get_base_dir();
    config.push(".config");
    let new_config = if !std::env::var("BORG_BASE_DIR").is_ok() {
        let mut new_config = match std::env::var("XDG_CONFIG_HOME") {
            Ok(p) => check_ok(&p),
            Err(_) => config,
        };
        new_config.push("borg");
        new_config
    } else {
        let config_path = match std::env::var("BORG_CONFIG_DIR") {
            Ok(p) => check_ok(&p),
            Err(_) => {
                config.push("borg");
                config
            }
        };
        config_path
    };
    if !new_config.is_dir() {
        eprintln!("{}: Failed to get borg config dir", &new_config.display());
        exit(1);
    }
    new_config
}

/// This function aims to get the borg security dir
///
/// # Description
/// Get the borg security folder defined by the BORG_SECURITY_DIR env variable
/// or get the borg_config_dir/security path
/// # Return
/// The path to  the borg security dir
fn get_security_dir() -> PathBuf {
    let security_dir = std::env::var("BORG_SECURITY_DIR");
    match security_dir {
        Ok(dir) => check_ok(&dir),
        Err(_) => {
            let mut path = get_config_dir();
            path.push("security");
            return path;
        }
    }
}

/// Get the timestamp manifest for the current borg project
/// # Arguments
/// - `security_dir`: The borg security folder
/// # Return
/// A file corresponding to the timestamp manifest of the project folder
fn get_manifest_timestamp(security_dir: PathBuf) -> PathBuf {
    let current_id = push::get_project_id();
    let mut security_dir = security_dir;
    security_dir.push(current_id);
    security_dir.push("manifest-timestamp");
    if !security_dir.is_file() {
        eprintln!(
            "{}: {} is not a file !",
            "error".red(),
            &security_dir.display()
        );
        exit(1);
    };
    security_dir
}

/// Function that remove the borg manifest-timestamp for the current borg project
fn remove_manifest_timestamp() {
    let security_dir = get_security_dir();
    let manifest = get_manifest_timestamp(security_dir);
    std::fs::remove_file(&manifest).expect(&format!(
        "{}: Unable to delete {}",
        "error".red(),
        &manifest.display()
    ));
}

/// Function that removes .borg cache
/// # Arguments
/// - `borg_folder`: The .borg folder of the current project
fn borg_delete_cache(borg_folder: &PathBuf) {
    let args = vec!["delete", "--cache-only", borg_folder.to_str().unwrap()];
    let mut output = Command::new("borg")
        .args(&args)
        .stdout(Stdio::inherit())
        .stdin(Stdio::inherit())
        .spawn()
        .unwrap();
    match output.wait() {
        Err(e) => {
            eprintln!("{}: borg terminaned with an error: {}", "error".red(), e);
            exit(1);
        }
        Ok(_) => (),
    }
}

/// Create a push the tar achive on the selected remote path
/// # Arguments
/// - `remote`: The name of a remote
pub fn pull(remote: &str) -> () {
    let (adress, url) = push::get_adress_and_url(remote);
    let (borg_folder, _) = commit::check_path();
    let borg_folder = borg_folder.canonicalize().unwrap();
    push::check_dest_and_copy(&borg_folder, &url, &adress);
    let remote_dir = push::get_remote_dir(&url);
    push::handle_existing_remote_dir(&remote_dir, &adress, "pull");
    save_and_clean_borg_folder(&borg_folder);
    push::copy_file(&borg_folder, &remote_dir, &adress, "pull");
    remove_manifest_timestamp();
    borg_delete_cache(&borg_folder);
}
