use crate::config_structure::{Config, RemoteConfig};
use crate::configt::{get_borgconfig, get_global_config, write_config};
use colored::Colorize;
use std::collections::BTreeMap as Dict;
use std::path::PathBuf;
use std::process::exit;

/// Get the configuration file of interest
/// # Argument
/// - global: A bolean indicating whether to get the global configuration or
///   only the local one
/// # Return
/// The path containing the local of global configuration or None
pub fn get_optional_config(global: bool) -> Option<PathBuf> {
    if global {
        get_global_config()
    } else {
        get_borgconfig()
    }
}

/// Return the global or the local configuration file parsed
/// # Arguments
/// # Return
/// The parsed global or local toml file
pub fn parse_config(global: bool) -> Config {
    let config_file = get_optional_config(global);
    let config_file = match config_file {
        None => {
            if global {
                return Config::new(global);
            } else {
                eprintln!("{}: local config should exists !", "error".red());
                exit(1);
            }
        }
        Some(c) => c,
    };
    let rcontent = std::fs::read_to_string(&config_file);
    let content = match rcontent {
        Err(e) => match e.raw_os_error() {
            Some(2) => String::from(""),
            _ => {
                eprintln!("Unable to read {} - {}", &config_file.to_str().unwrap(), e);
                exit(101);
            }
        },
        Ok(s) => s,
    };
    if content.is_empty() {
        if global {
            return Config::new(global);
        } else {
            eprintln!("{}: local config should not be empty !", "error".red());
            exit(1);
        }
    }
    let config_str: Config = match toml::from_str(&content) {
        Ok(config) => config,
        Err(_) => {
            eprintln!(
                "{}: bad {} configuration file !",
                "error".red(),
                &config_file.display()
            );
            exit(1);
        }
    };
    config_str.check_project_type(global, &config_file);
    config_str
}

/// Displays given in input
/// # Arguments
/// - `remotes`: A dictionary linking  remote name and remote config
fn display_remotes(remotes: Dict<String, (RemoteConfig, String)>) -> () {
    for (_, (rconfig, tag)) in remotes {
        rconfig.show(&tag);
    }
}

/// Combine remote before printing them
/// # Arguments
/// - global: Global remote config
/// - local: Local remote config
/// # Return:
/// A dictionary linking  remote name and remote config
fn combine_remote(
    global: Option<Vec<RemoteConfig>>,
    local: Option<Vec<RemoteConfig>>,
) -> Dict<String, (RemoteConfig, String)> {
    let glo = match global {
        Some(v) => v,
        None => Vec::new(),
    };
    let loc = match local {
        Some(v) => v,
        None => Vec::new(),
    };
    let mut dic: Dict<String, (RemoteConfig, String)> = Dict::new();
    for l in loc.into_iter() {
        dic.insert(l.name.to_owned(), (l, String::from("")));
    }
    for g in glo.into_iter() {
        dic.entry(g.name.to_owned())
            .or_insert((g, String::from("(global)")));
    }
    dic
}

/// Get the remote config of a configuration file
/// # Arguments
/// - global: a boolean indicating to returns the remotes the the global or
///   local configuration file
/// # Return
/// Return the remotes of the local or the global configuration file
fn get_remotes(global: bool) -> Option<Vec<RemoteConfig>> {
    let my_config = parse_config(global);
    my_config.gblk_remote
}

/// Get a dictionary of locally and globally defined remotes
/// # Return
/// A dictionary linking remote name and remote config globally and locally
pub(crate) fn get_dic_remotes() -> Dict<String, (RemoteConfig, String)> {
    let remote_global = get_remotes(true);
    let remote_local = get_remotes(false);
    combine_remote(remote_global, remote_local)
}

/// Function that show the current gblk config of the borg folder
pub fn show() -> () {
    let dic_remote = get_dic_remotes();
    display_remotes(dic_remote);
}

/// Function that update the current gblk config file for the project folder
/// # Arguments
/// - `key` : The remote name to update
/// - `value`: The value of the key to update
/// - `global: A bolean indicating whether to update the  global or the local configuration
pub(crate) fn update_config(key: &str, value: &str, global: bool) -> () {
    let mut my_config = parse_config(global);
    let mut remotes = match my_config.gblk_remote {
        Some(r) => r,
        None => Vec::new(),
    };
    let mut updated = false;
    for r in remotes.iter_mut() {
        if r.name == key {
            r.url = value.to_owned();
            updated = true;
        }
    }
    if !updated {
        remotes.push(RemoteConfig {
            name: key.to_owned(),
            url: value.to_owned(),
        })
    }
    my_config.gblk_remote = Some(remotes);
    write_config(&my_config, global);
}

/// Function that remove a given remote in the current gblk config file for the project folder
/// # Arguments
/// - `key` : The remote name to delete
/// - `global: A bolean indicating whether to remove the key in the  global or
///   the local configuration file
pub fn remove_config(key: &str, global: bool) -> () {
    let mut my_config = parse_config(global);
    let mut remotes = match my_config.gblk_remote {
        Some(r) => r,
        None => Vec::new(),
    };
    let init_size = remotes.len();
    remotes.retain(|x| &x.name != &key);
    let end_size = remotes.len();
    if init_size == end_size {
        eprintln!("{}: The remote {} doesn't exists !", "error".red(), key);
        exit(1);
    }
    my_config.gblk_remote = if remotes.len() > 0 {
        Some(remotes)
    } else {
        None
    };
    write_config(&my_config, global);
}
